from . import plot
#from . import plot_legacy
#from . import plot_widgets
from . import XYplotWidget
from . import XYPlotDevice
from . import rawdatadisp
from . import PcolorPlotDevice
redvypr_devicemodule = True